@@ -1,11 +1,12 @@
 #!/bin/sh
 #
-# $Id: 1_prepare_files.sh,v 1.11 2004/05/25 04:22:49 nigel Exp $
+# $Id: 1_prepare_files.sh,v 1.13 2005/12/29 13:06:12 nigel Exp $
 #
 # Run this to generate all the initial makefiles, etc.
 
 ln -sf ../Unix/config.guess .
 ln -sf ../Unix/config.sub   .
+ln -sf ../Unix/semaphore.h  .
 cp -pf ../Unix/sys_unix.cpp .
 cp -pf ../Unix/timer_unix.cpp .
 ln -sf ../Unix/user_strings_unix.h .
@@ -50,8 +51,4 @@ fi
 # This mode isn't working yet - segfaults
 #./configure "$@" --enable-addressing=real
 
-# The default mode, which doesn't allow Classic ROMS and has black screen prob.
-#./configure "$@"
-
-# Classic ROM version, seems to fix black screen problem
-./configure "$@" --enable-addressing=banks
+./configure "$@"
