@@ -86,9 +86,10 @@ using std::string;
 #include "main.h"
 #include "vm_alloc.h"
 #include "sigsegv.h"
+#include "rpc.h"
 
 #if USE_JIT
-extern void flush_icache_range(uint32 start, uint32 size); // from compemu_support.cpp
+extern void flush_icache_range(uint8 *start, uint32 size); // from compemu_support.cpp
 #endif
 
 #ifdef ENABLE_MON
@@ -192,6 +193,9 @@ static void sigint_handler(...);
 static bool lm_area_mapped = false;	// Flag: Low Memory area mmap()ped
 #endif
 
+static rpc_connection_t *gui_connection = NULL;	// RPC connection to the GUI
+static const char *gui_connection_path = NULL;	// GUI connection identifier
+
 
 // Prototypes
 static void *xpram_func(void *arg);
@@ -223,19 +227,36 @@ char *strdup(const char *s)
 
 
 /*
- *  Map memory that can be accessed from the Mac side
+ *  Helpers to map memory that can be accessed from the Mac side
  */
 
+// NOTE: VM_MAP_33BIT is only used when compiling a 64-bit JIT on specific platforms
 void *vm_acquire_mac(size_t size)
 {
 	void *m = vm_acquire(size, VM_MAP_DEFAULT | VM_MAP_33BIT);
-	if (m == NULL) {
+#ifdef USE_33BIT_ADDRESSING
+	if (m == VM_MAP_FAILED) {
+		printf("WARNING: Cannot acquire memory in 33-bit address space (%s)\n", strerror(errno));
 		ThirtyThreeBitAddressing = false;
 		m = vm_acquire(size);
 	}
+#endif
 	return m;
 }
 
+static int vm_acquire_mac_fixed(void *addr, size_t size)
+{
+	int ret = vm_acquire_fixed(addr, size, VM_MAP_DEFAULT | VM_MAP_33BIT);
+#ifdef USE_33BIT_ADDRESSING
+	if (ret < 0) {
+		printf("WARNING: Cannot acquire fixed memory in 33-bit address space (%s)\n", strerror(errno));
+		ThirtyThreeBitAddressing = false;
+		ret = vm_acquire_fixed(addr, size);
+	}
+#endif
+	return ret;
+}
+
 
 /*
  *  SIGSEGV handler
@@ -396,6 +417,12 @@ int main(int argc, char **argv)
 			if (i < argc)
 				x_display_name = strdup(argv[i]);
 #endif
+		} else if (strcmp(argv[i], "--gui-connection") == 0) {
+			argv[i++] = NULL;
+			if (i < argc) {
+				gui_connection_path = argv[i];
+				argv[i] = NULL;
+			}
 		} else if (strcmp(argv[i], "--break") == 0) {
 			argv[i++] = NULL;
 			if (i < argc) {
@@ -429,17 +456,27 @@ int main(int argc, char **argv)
 		}
 	}
 
+	// Connect to the external GUI
+	if (gui_connection_path) {
+		if ((gui_connection = rpc_init_client(gui_connection_path)) == NULL) {
+			fprintf(stderr, "Failed to initialize RPC client connection to the GUI\n");
+			return 1;
+		}
+	}
+
 #ifdef ENABLE_GTK
+	if (!gui_connection) {
 #ifdef HAVE_GNOMEUI
-	// Init GNOME/GTK
-	char version[16];
-	sprintf(version, "%d.%d", VERSION_MAJOR, VERSION_MINOR);
-	gnome_init("Basilisk II", version, argc, argv);
+		// Init GNOME/GTK
+		char version[16];
+		sprintf(version, "%d.%d", VERSION_MAJOR, VERSION_MINOR);
+		gnome_init("Basilisk II", version, argc, argv);
 #else
-	// Init GTK
-	gtk_set_locale();
-	gtk_init(&argc, &argv);
+		// Init GTK
+		gtk_set_locale();
+		gtk_init(&argc, &argv);
 #endif
+	}
 #endif
 
 	// Read preferences
@@ -492,7 +529,7 @@ int main(int argc, char **argv)
 	SysInit();
 
 	// Show preferences editor
-	if (!PrefsFindBool("nogui"))
+	if (!gui_connection && !PrefsFindBool("nogui"))
 		if (!PrefsEditor())
 			QuitEmulator();
 
@@ -512,6 +549,8 @@ int main(int argc, char **argv)
 		WarningAlert(GetString(STR_SMALL_RAM_WARN));
 		RAMSize = 1024*1024;
 	}
+	if (RAMSize > 1023*1024*1024)						// Cap to 1023MB (APD crashes at 1GB)
+		RAMSize = 1023*1024*1024;
 
 #if REAL_ADDRESSING || DIRECT_ADDRESSING
 	RAMSize = RAMSize & -getpagesize();					// Round down to page boundary
@@ -520,27 +559,33 @@ int main(int argc, char **argv)
 	// Initialize VM system
 	vm_init();
 
+#ifdef USE_33BIT_ADDRESSING
+	// Speculatively enables 33-bit addressing
+	ThirtyThreeBitAddressing = true;
+#endif
+
 #if REAL_ADDRESSING
 	// Flag: RAM and ROM are contigously allocated from address 0
 	bool memory_mapped_from_zero = false;
-	
-	// Under Solaris/SPARC and NetBSD/m68k, Basilisk II is known to crash
-	// when trying to map a too big chunk of memory starting at address 0
-#if defined(OS_solaris) || defined(OS_netbsd) || defined(PAGEZERO_HACK)
-	const bool can_map_all_memory = false;
-#else
+
+	// Make sure to map RAM & ROM at address 0 only on platforms that
+	// supports linker scripts to relocate the Basilisk II executable
+	// above 0x70000000
+#if HAVE_LINKER_SCRIPT
 	const bool can_map_all_memory = true;
+#else
+	const bool can_map_all_memory = false;
 #endif
 	
 	// Try to allocate all memory from 0x0000, if it is not known to crash
-	if (can_map_all_memory && (vm_acquire_fixed(0, RAMSize + 0x100000) == 0)) {
+	if (can_map_all_memory && (vm_acquire_mac_fixed(0, RAMSize + 0x100000) == 0)) {
 		D(bug("Could allocate RAM and ROM from 0x0000\n"));
 		memory_mapped_from_zero = true;
 	}
 	
 #ifndef PAGEZERO_HACK
 	// Otherwise, just create the Low Memory area (0x0000..0x2000)
-	else if (vm_acquire_fixed(0, 0x2000) == 0) {
+	else if (vm_acquire_mac_fixed(0, 0x2000) == 0) {
 		D(bug("Could allocate the Low Memory globals\n"));
 		lm_area_mapped = true;
 	}
@@ -563,10 +608,6 @@ int main(int argc, char **argv)
 	else
 #endif
 	{
-#ifdef USE_33BIT_ADDRESSING
-		// Speculatively enables 33-bit addressing
-		ThirtyThreeBitAddressing = true;
-#endif
 		uint8 *ram_rom_area = (uint8 *)vm_acquire_mac(RAMSize + 0x100000);
 		if (ram_rom_area == VM_MAP_FAILED) {	
 			ErrorAlert(STR_NO_MEM_ERR);
@@ -578,7 +619,7 @@ int main(int argc, char **argv)
 
 #if USE_SCRATCHMEM_SUBTERFUGE
 	// Allocate scratch memory
-	ScratchMem = (uint8 *)vm_acquire(SCRATCH_MEM_SIZE);
+	ScratchMem = (uint8 *)vm_acquire_mac(SCRATCH_MEM_SIZE);
 	if (ScratchMem == VM_MAP_FAILED) {
 		ErrorAlert(STR_NO_MEM_ERR);
 		QuitEmulator();
@@ -593,8 +634,8 @@ int main(int argc, char **argv)
 	ROMBaseMac = Host2MacAddr(ROMBaseHost);
 #endif
 #if REAL_ADDRESSING
-	RAMBaseMac = (uint32)RAMBaseHost;
-	ROMBaseMac = (uint32)ROMBaseHost;
+	RAMBaseMac = Host2MacAddr(RAMBaseHost);
+	ROMBaseMac = Host2MacAddr(ROMBaseHost);
 #endif
 	D(bug("Mac RAM starts at %p (%08x)\n", RAMBaseHost, RAMBaseMac));
 	D(bug("Mac ROM starts at %p (%08x)\n", ROMBaseHost, ROMBaseMac));
@@ -885,6 +926,12 @@ void QuitEmulator(void)
 		XCloseDisplay(x_display);
 #endif
 
+	// Notify GUI we are about to leave
+	if (gui_connection) {
+		if (rpc_method_invoke(gui_connection, RPC_METHOD_EXIT, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR)
+			rpc_method_wait_for_reply(gui_connection, RPC_TYPE_INVALID);
+	}
+
 	exit(0);
 }
 
@@ -898,7 +945,7 @@ void FlushCodeCache(void *start, uint32 size)
 {
 #if USE_JIT
     if (UseJIT)
-		flush_icache_range((uintptr)start, size);
+		flush_icache_range((uint8 *)start, size);
 #endif
 #if !EMULATED_68K && defined(__NetBSD__)
 	m68k_sync_icache(start, size);
@@ -1525,6 +1572,11 @@ void display_alert(int title_id, int prefix_id, int button_id, const char *text)
 
 void ErrorAlert(const char *text)
 {
+	if (gui_connection) {
+		if (rpc_method_invoke(gui_connection, RPC_METHOD_ERROR_ALERT, RPC_TYPE_STRING, text, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR &&
+			rpc_method_wait_for_reply(gui_connection, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR)
+			return;
+	}
 #if defined(ENABLE_GTK) && !defined(USE_SDL_VIDEO)
 	if (PrefsFindBool("nogui") || x_display == NULL) {
 		printf(GetString(STR_SHELL_ERROR_PREFIX), text);
@@ -1544,6 +1596,11 @@ void ErrorAlert(const char *text)
 
 void WarningAlert(const char *text)
 {
+	if (gui_connection) {
+		if (rpc_method_invoke(gui_connection, RPC_METHOD_WARNING_ALERT, RPC_TYPE_STRING, text, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR &&
+			rpc_method_wait_for_reply(gui_connection, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR)
+			return;
+	}
 #if defined(ENABLE_GTK) && !defined(USE_SDL_VIDEO)
 	if (PrefsFindBool("nogui") || x_display == NULL) {
 		printf(GetString(STR_SHELL_WARNING_PREFIX), text);
