@@ -599,15 +599,22 @@ if (ret != KERN_SUCCESS) { \
 #define SIGSEGV_REGISTER_FILE			(unsigned long *)&state->srr0, (unsigned long *)&state->r0
 #endif
 #ifdef __i386__
+#ifdef i386_SAVED_STATE
 #define SIGSEGV_THREAD_STATE_TYPE		struct i386_saved_state
 #define SIGSEGV_THREAD_STATE_FLAVOR		i386_SAVED_STATE
 #define SIGSEGV_THREAD_STATE_COUNT		i386_SAVED_STATE_COUNT
+#define SIGSEGV_REGISTER_FILE			((unsigned long *)&state->edi) /* EDI is the first GPR we consider */
+#else
+#define SIGSEGV_THREAD_STATE_TYPE		struct i386_thread_state
+#define SIGSEGV_THREAD_STATE_FLAVOR		i386_THREAD_STATE
+#define SIGSEGV_THREAD_STATE_COUNT		i386_THREAD_STATE_COUNT
+#define SIGSEGV_REGISTER_FILE			((unsigned long *)&state->eax) /* EAX is the first GPR we consider */
+#endif
 #define SIGSEGV_FAULT_INSTRUCTION		state->eip
 #define SIGSEGV_SKIP_INSTRUCTION		ix86_skip_instruction
-#define SIGSEGV_REGISTER_FILE			((unsigned long *)&state->edi) /* EDI is the first GPR we consider */
 #endif
 #define SIGSEGV_FAULT_ADDRESS			code[1]
-#define SIGSEGV_FAULT_HANDLER_INVOKE(ADDR, IP)	((code[0] == KERN_PROTECTION_FAILURE) ? sigsegv_fault_handler(ADDR, IP) : SIGSEGV_RETURN_FAILURE)
+#define SIGSEGV_FAULT_HANDLER_INVOKE(ADDR, IP)	((code[0] == KERN_PROTECTION_FAILURE || code[0] == KERN_INVALID_ADDRESS) ? sigsegv_fault_handler(ADDR, IP) : SIGSEGV_RETURN_FAILURE)
 #define SIGSEGV_FAULT_HANDLER_ARGLIST	mach_port_t thread, exception_data_t code, SIGSEGV_THREAD_STATE_TYPE *state
 #define SIGSEGV_FAULT_HANDLER_ARGS		thread, code, &state
 
@@ -766,8 +773,9 @@ enum {
 };
 #endif
 #if defined(__APPLE__) && defined(__MACH__)
-// expected to be the same as FreeBSD
 enum {
+#ifdef i386_SAVED_STATE
+	// same as FreeBSD (in Open Darwin 8.0.1)
 	X86_REG_EIP = 10,
 	X86_REG_EAX = 7,
 	X86_REG_ECX = 6,
@@ -777,6 +785,18 @@ enum {
 	X86_REG_EBP = 2,
 	X86_REG_ESI = 1,
 	X86_REG_EDI = 0
+#else
+	// new layout (MacOS X 10.4.4 for x86)
+	X86_REG_EIP = 10,
+	X86_REG_EAX = 0,
+	X86_REG_ECX = 2,
+	X86_REG_EDX = 4,
+	X86_REG_EBX = 1,
+	X86_REG_ESP = 7,
+	X86_REG_EBP = 6,
+	X86_REG_ESI = 5,
+	X86_REG_EDI = 4
+#endif
 };
 #endif
 #if defined(_WIN32)
@@ -923,6 +943,33 @@ static bool ix86_skip_instruction(unsigned long * regs)
 			break;
 	    }
 	  break;
+#if defined(__x86_64__)
+	case 0x63: // MOVSXD r64, r/m32
+		if (has_rex && rex.W) {
+			transfer_size = SIZE_LONG;
+			target_size = SIZE_QUAD;
+		}
+		else if (transfer_size != SIZE_WORD) {
+			transfer_size = SIZE_LONG;
+			target_size = SIZE_QUAD;
+		}
+		switch (eip[1] & 0xc0) {
+		case 0x80:
+			reg = (eip[1] >> 3) & 7;
+			transfer_type = SIGSEGV_TRANSFER_LOAD;
+			break;
+		case 0x40:
+			reg = (eip[1] >> 3) & 7;
+			transfer_type = SIGSEGV_TRANSFER_LOAD;
+			break;
+		case 0x00:
+			reg = (eip[1] >> 3) & 7;
+			transfer_type = SIGSEGV_TRANSFER_LOAD;
+			break;
+		}
+		len += 2 + ix86_step_over_modrm(eip + 1);
+		break;
+#endif
 	case 0x8a: // MOV r8, r/m8
 		transfer_size = SIZE_BYTE;
 	case 0x8b: // MOV r32, r/m32 (or 16-bit operation)
@@ -1271,6 +1318,7 @@ enum {
   SPARC_REG_G1 = REG_G1,
   SPARC_REG_O0 = REG_O0,
   SPARC_REG_PC = REG_PC,
+  SPARC_REG_nPC = REG_nPC
 #endif
 };
 static bool sparc_skip_instruction(unsigned long * regs, gwindows_t * gwins, struct rwindow * rwin)
@@ -1334,7 +1382,7 @@ static bool sparc_skip_instruction(unsigned long * regs, gwindows_t * gwins, str
 	break;
   case 7: // Store Doubleword
 	transfer_type = SIGSEGV_TRANSFER_STORE;
-	transfer_size = SIZE_WORD;
+	transfer_size = SIZE_LONG;
 	register_pair = true;
 	break;
   }
@@ -1344,33 +1392,7 @@ static bool sparc_skip_instruction(unsigned long * regs, gwindows_t * gwins, str
 	return false;
   }
 
-  // Zero target register in case of a load operation
   const int reg = (opcode >> 25) & 0x1f;
-  if (transfer_type == SIGSEGV_TRANSFER_LOAD && reg != 0) {
-	// FIXME: code to handle local & input registers is not tested
-	if (reg >= 1 && reg <= 7) {
-	  // global registers
-	  regs[reg - 1 + SPARC_REG_G1] = 0;
-	}
-	else if (reg >= 8 && reg <= 15) {
-	  // output registers
-	  regs[reg - 8 + SPARC_REG_O0] = 0;
-	}
-	else if (reg >= 16 && reg <= 23) {
-	  // local registers (in register windows)
-	  if (gwins)
-		gwins->wbuf->rw_local[reg - 16] = 0;
-	  else
-		rwin->rw_local[reg - 16] = 0;
-	}
-	else {
-	  // input registers (in register windows)
-	  if (gwins)
-		gwins->wbuf->rw_in[reg - 24] = 0;
-	  else
-		rwin->rw_in[reg - 24] = 0;
-	}
-  }
 
 #if DEBUG
   static const char * reg_names[] = {
@@ -1388,7 +1410,35 @@ static bool sparc_skip_instruction(unsigned long * regs, gwindows_t * gwins, str
 		 reg_names[reg]);
 #endif
 
+  // Zero target register in case of a load operation
+  if (transfer_type == SIGSEGV_TRANSFER_LOAD && reg != 0) {
+	// FIXME: code to handle local & input registers is not tested
+	if (reg >= 1 && reg < 8) {
+	  // global registers
+	  regs[reg - 1 + SPARC_REG_G1] = 0;
+	}
+	else if (reg >= 8 && reg < 16) {
+	  // output registers
+	  regs[reg - 8 + SPARC_REG_O0] = 0;
+	}
+	else if (reg >= 16 && reg < 24) {
+	  // local registers (in register windows)
+	  if (gwins)
+		gwins->wbuf->rw_local[reg - 16] = 0;
+	  else
+		rwin->rw_local[reg - 16] = 0;
+	}
+	else {
+	  // input registers (in register windows)
+	  if (gwins)
+		gwins->wbuf->rw_in[reg - 24] = 0;
+	  else
+		rwin->rw_in[reg - 24] = 0;
+	}
+  }
+
   regs[SPARC_REG_PC] += 4;
+  regs[SPARC_REG_nPC] += 4;
   return true;
 }
 #endif
@@ -1650,7 +1700,7 @@ forward_exception(mach_port_t thread_port,
 	mach_port_t port;
 	exception_behavior_t behavior;
 	thread_state_flavor_t flavor;
-	thread_state_t thread_state;
+	thread_state_data_t thread_state;
 	mach_msg_type_number_t thread_state_count;
 
 	for (portIndex = 0; portIndex < oldExceptionPorts->maskCount; portIndex++) {
@@ -1669,13 +1719,18 @@ forward_exception(mach_port_t thread_port,
 	behavior = oldExceptionPorts->behaviors[portIndex];
 	flavor = oldExceptionPorts->flavors[portIndex];
 
+	if (!VALID_THREAD_STATE_FLAVOR(flavor)) {
+		fprintf(stderr, "Invalid thread_state flavor = %d. Not forwarding\n", flavor);
+		return KERN_FAILURE;
+	}
+
 	/*
 	 fprintf(stderr, "forwarding exception, port = 0x%x, behaviour = %d, flavor = %d\n", port, behavior, flavor);
 	 */
 
 	if (behavior != EXCEPTION_DEFAULT) {
 		thread_state_count = THREAD_STATE_MAX;
-		kret = thread_get_state (thread_port, flavor, thread_state,
+		kret = thread_get_state (thread_port, flavor, (natural_t *)&thread_state,
 								 &thread_state_count);
 		MACH_CHECK_ERROR (thread_get_state, kret);
 	}
@@ -1691,8 +1746,8 @@ forward_exception(mach_port_t thread_port,
 	  // fprintf(stderr, "forwarding to exception_raise_state\n");
 	  kret = exception_raise_state(port, exception_type, exception_data,
 								   data_count, &flavor,
-								   thread_state, thread_state_count,
-								   thread_state, &thread_state_count);
+								   (natural_t *)&thread_state, thread_state_count,
+								   (natural_t *)&thread_state, &thread_state_count);
 	  MACH_CHECK_ERROR (exception_raise_state, kret);
 	  break;
 	case EXCEPTION_STATE_IDENTITY:
@@ -1700,22 +1755,23 @@ forward_exception(mach_port_t thread_port,
 	  kret = exception_raise_state_identity(port, thread_port, task_port,
 											exception_type, exception_data,
 											data_count, &flavor,
-											thread_state, thread_state_count,
-											thread_state, &thread_state_count);
+											(natural_t *)&thread_state, thread_state_count,
+											(natural_t *)&thread_state, &thread_state_count);
 	  MACH_CHECK_ERROR (exception_raise_state_identity, kret);
 	  break;
 	default:
 	  fprintf(stderr, "forward_exception got unknown behavior\n");
+	  kret = KERN_FAILURE;
 	  break;
 	}
 
 	if (behavior != EXCEPTION_DEFAULT) {
-		kret = thread_set_state (thread_port, flavor, thread_state,
+		kret = thread_set_state (thread_port, flavor, (natural_t *)&thread_state,
 								 thread_state_count);
 		MACH_CHECK_ERROR (thread_set_state, kret);
 	}
 
-	return KERN_SUCCESS;
+	return kret;
 }
 
 /*
@@ -2068,6 +2124,13 @@ static int page_size;
 static volatile char * page = 0;
 static volatile int handler_called = 0;
 
+/* Barriers */
+#ifdef __GNUC__
+#define BARRIER() asm volatile ("" : : : "memory")
+#else
+#define BARRIER() /* nothing */
+#endif
+
 #ifdef __GNUC__
 // Code range where we expect the fault to come from
 static void *b_region, *e_region;
@@ -2159,6 +2222,8 @@ static bool arch_insn_skipper_tests()
 		0x4c, 0x89, 0x18,              // mov    %r11,(%rax)
 		0x4a, 0x89, 0x0c, 0x10,        // mov    %rcx,(%rax,%r10,1)
 		0x4e, 0x89, 0x1c, 0x10,        // mov    %r11,(%rax,%r10,1)
+		0x63, 0x47, 0x04,              // movslq 4(%rdi),%eax
+		0x48, 0x63, 0x47, 0x04,        // movslq 4(%rdi),%rax
 #endif
 		0                              // end
 	};
@@ -2202,6 +2267,7 @@ int main(void)
 	if (page[REF_INDEX] != REF_VALUE)
 	  exit(20);
 	page[REF_INDEX] = REF_VALUE;
+	BARRIER();
  L_e_region1:
 
 	if (handler_called != 1)
@@ -2242,6 +2308,7 @@ int main(void)
 	TEST_SKIP_INSTRUCTION(signed short);
 	TEST_SKIP_INSTRUCTION(signed int);
 	TEST_SKIP_INSTRUCTION(signed long);
+	BARRIER();
  L_e_region2:
 
 	if (!arch_insn_skipper_tests())
