@@ -20,11 +20,13 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#include "sysdeps.h"
+
 #include <process.h>
 #include <windowsx.h>
+#include <winioctl.h>
 #include <ctype.h>
 
-#include "sysdeps.h"
 #include "cpu_emulation.h"
 #include "main.h"
 #include "macos_util.h"
@@ -37,6 +39,10 @@
 #include "ether_windows.h"
 #include "router/router.h"
 #include "kernel_windows.h"
+#include "libslirp.h"
+
+// Define to let the slirp library determine the right timeout for select()
+#define USE_SLIRP_TIMEOUT 1
 
 
 #define DEBUG 0
@@ -49,6 +55,37 @@
 #include "debug.h"
 
 
+// Ethernet device types
+enum {
+	NET_IF_B2ETHER,
+	NET_IF_ROUTER,
+	NET_IF_SLIRP,
+	NET_IF_TAP,
+	NET_IF_FAKE,
+};
+
+// TAP-Win32 constants
+#define TAP_VERSION_MIN_MAJOR 7
+#define TAP_VERSION_MIN_MINOR 1
+
+#define TAP_CONTROL_CODE(request, method) \
+		CTL_CODE (FILE_DEVICE_UNKNOWN, request, method, FILE_ANY_ACCESS)
+
+#define TAP_IOCTL_GET_MAC				TAP_CONTROL_CODE (1, METHOD_BUFFERED)
+#define TAP_IOCTL_GET_VERSION			TAP_CONTROL_CODE (2, METHOD_BUFFERED)
+#define TAP_IOCTL_GET_MTU				TAP_CONTROL_CODE (3, METHOD_BUFFERED)
+#define TAP_IOCTL_GET_INFO				TAP_CONTROL_CODE (4, METHOD_BUFFERED)
+#define TAP_IOCTL_CONFIG_POINT_TO_POINT	TAP_CONTROL_CODE (5, METHOD_BUFFERED)
+#define TAP_IOCTL_SET_MEDIA_STATUS		TAP_CONTROL_CODE (6, METHOD_BUFFERED)
+#define TAP_IOCTL_CONFIG_DHCP_MASQ		TAP_CONTROL_CODE (7, METHOD_BUFFERED)
+#define TAP_IOCTL_GET_LOG_LINE			TAP_CONTROL_CODE (8, METHOD_BUFFERED)
+#define TAP_IOCTL_CONFIG_DHCP_SET_OPT	TAP_CONTROL_CODE (9, METHOD_BUFFERED)
+
+#define OLD_TAP_CONTROL_CODE(request, method) \
+		CTL_CODE (FILE_DEVICE_PHYSICAL_NETCARD | 8000, request, method, FILE_ANY_ACCESS)
+
+#define OLD_TAP_IOCTL_GET_VERSION		OLD_TAP_CONTROL_CODE (3, METHOD_BUFFERED)
+
 // Options
 bool ether_use_permanent = true;
 static int16 ether_multi_mode = ETHER_MULTICAST_MAC;
@@ -58,20 +95,21 @@ HANDLE ether_th;
 unsigned int ether_tid;
 HANDLE ether_th1;
 HANDLE ether_th2;
-
-
-// Need to fake a NIC if there is none but the router module is activated.
-bool ether_fake = false;
+static int net_if_type = -1;	// Ethernet device type
+#ifdef SHEEPSHAVER
+static bool net_open = false;	// Flag: initialization succeeded, network device open
+uint8 ether_addr[6];			// Our Ethernet address
+#endif
 
 // These are protected by queue_csection
 // Controls transfer for read thread to feed thread
 static CRITICAL_SECTION queue_csection;
-typedef struct _queue_t {
+typedef struct _win_queue_t {
 	uint8 *buf;
 	int sz;
-} queue_t;
+} win_queue_t;
 #define MAX_QUEUE_ITEMS 1024
-static queue_t queue[MAX_QUEUE_ITEMS];
+static win_queue_t queue[MAX_QUEUE_ITEMS];
 static int queue_head = 0;
 static int queue_inx = 0;
 static bool wait_request = true;
@@ -129,10 +167,15 @@ static HANDLE int_sig = 0;
 static HANDLE int_sig2 = 0;
 static HANDLE int_send_now = 0;
 
-static char edevice[512];
-
-
 // Prototypes
+static LPADAPTER tap_open_adapter(const char *dev_name);
+static void tap_close_adapter(LPADAPTER fd);
+static bool tap_check_version(LPADAPTER fd);
+static bool tap_set_status(LPADAPTER fd, ULONG status);
+static bool tap_get_mac(LPADAPTER fd, LPBYTE addr);
+static bool tap_receive_packet(LPADAPTER fd, LPPACKET lpPacket, BOOLEAN Sync);
+static bool tap_send_packet(LPADAPTER fd, LPPACKET lpPacket, BOOLEAN Sync, BOOLEAN recycle);
+static WINAPI unsigned int slirp_receive_func(void *arg);
 static WINAPI unsigned int ether_thread_feed_int(void *arg);
 static WINAPI unsigned int ether_thread_get_packets_nt(void *arg);
 static WINAPI unsigned int ether_thread_write_packets(void *arg);
@@ -141,6 +184,10 @@ static void final_queue(void);
 static bool allocate_read_packets(void);
 static void free_read_packets(void);
 static void free_write_packets(void);
+static int16 ether_do_add_multicast(uint8 *addr);
+static int16 ether_do_del_multicast(uint8 *addr);
+static int16 ether_do_write(uint32 arg);
+static void ether_do_interrupt(void);
 
 
 /*
@@ -173,47 +220,72 @@ bool ether_init(void)
 {
 	char str[256];
 
-	// Initialize NAT-Router
-	router_init();
-
 	// Do nothing if no Ethernet device specified
 	const char *name = PrefsFindString("ether");
-	if (name)
-		strcpy(edevice, name);
+	if (name == NULL)
+		return false;
 
-	bool there_is_a_router = PrefsFindBool("routerenabled");
+	ether_multi_mode = PrefsFindInt32("ethermulticastmode");
+	ether_use_permanent = PrefsFindBool("etherpermanentaddress");
 
-	if (!name || !*name) {
-		if( there_is_a_router ) {
-			strcpy( edevice, "None" );
-			ether_fake = true;
-		} else {
+	// Determine Ethernet device type
+	net_if_type = -1;
+	if (PrefsFindBool("routerenabled") || strcmp(name, "router") == 0)
+		net_if_type = NET_IF_ROUTER;
+	else if (strcmp(name, "slirp") == 0)
+		net_if_type = NET_IF_SLIRP;
+	else if (strcmp(name, "tap") == 0)
+		net_if_type = NET_IF_TAP;
+	else
+		net_if_type = NET_IF_B2ETHER;
+
+	// Initialize NAT-Router
+	if (net_if_type == NET_IF_ROUTER) {
+		if (!router_init())
+			net_if_type = NET_IF_FAKE;
+	}
+
+	// Initialize slirp library
+	if (net_if_type == NET_IF_SLIRP) {
+		if (slirp_init() < 0) {
+			sprintf(str, GetString(STR_SLIRP_NO_DNS_FOUND_WARN));
+			WarningAlert(str);
 			return false;
 		}
 	}
 
-	ether_use_permanent = PrefsFindBool("etherpermanentaddress");
-	ether_multi_mode = PrefsFindInt32("ethermulticastmode");
-
 	// Open ethernet device
-	if(ether_fake) {
-		memcpy( ether_addr, router_mac_addr, 6 );
-		D(bug("Fake ethernet address (same as router) %02x %02x %02x %02x %02x %02x\r\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
-	} else {
-		fd = PacketOpenAdapter( name, ether_multi_mode );
+	const char *dev_name;
+	switch (net_if_type) {
+	case NET_IF_B2ETHER:
+		dev_name = PrefsFindString("etherguid");
+		if (dev_name == NULL || strcmp(name, "b2ether") != 0)
+			dev_name = name;
+		break;
+	case NET_IF_TAP:
+		dev_name = PrefsFindString("etherguid");
+		break;
+	}
+	if (net_if_type == NET_IF_B2ETHER) {
+		if (dev_name == NULL) {
+			WarningAlert("No ethernet device GUID specified. Ethernet is not available.");
+			goto open_error;
+		}
+
+		fd = PacketOpenAdapter( dev_name, ether_multi_mode );
 		if (!fd) {
-			sprintf(str, "Could not open ethernet adapter %s.", name);
+			sprintf(str, "Could not open ethernet adapter %s.", dev_name);
 			WarningAlert(str);
 			goto open_error;
 		}
 
 		// Get Ethernet address
 		if(!PacketGetMAC(fd,ether_addr,ether_use_permanent)) {
-			sprintf(str, "Could not get hardware address of device %s. Ethernet is not available.", name);
+			sprintf(str, "Could not get hardware address of device %s. Ethernet is not available.", dev_name);
 			WarningAlert(str);
 			goto open_error;
 		}
-		D(bug("Real ethernet address %02x %02x %02x %02x %02x %02x\r\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
+		D(bug("Real ethernet address %02x %02x %02x %02x %02x %02x\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
 
 		const char *ether_fake_address;
 		ether_fake_address = PrefsFindString("etherfakeaddress");
@@ -225,9 +297,78 @@ bool ether_init(void)
 				sm[3] = ether_fake_address[i*2+1];
 				ether_addr[i] = (uint8)strtoul(sm,0,0);
 			}
-			D(bug("Fake ethernet address %02x %02x %02x %02x %02x %02x\r\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
+			D(bug("Fake ethernet address %02x %02x %02x %02x %02x %02x\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
 		}
 	}
+	else if (net_if_type == NET_IF_TAP) {
+		if (dev_name == NULL) {
+			WarningAlert("No ethernet device GUID specified. Ethernet is not available.");
+			goto open_error;
+		}
+
+		fd = tap_open_adapter(dev_name);
+		if (!fd) {
+			sprintf(str, "Could not open ethernet adapter %s.", dev_name);
+			WarningAlert(str);
+			goto open_error;
+		}
+
+		if (!tap_check_version(fd)) {
+			sprintf(str, "Minimal TAP-Win32 version supported is %d.%d.", TAP_VERSION_MIN_MAJOR, TAP_VERSION_MIN_MINOR);
+			WarningAlert(str);
+			goto open_error;
+		}
+
+		if (!tap_set_status(fd, true)) {
+			sprintf(str, "Could not set media status to connected.");
+			WarningAlert(str);
+			goto open_error;
+		}
+
+		if (!tap_get_mac(fd, ether_addr)) {
+			sprintf(str, "Could not get hardware address of device %s. Ethernet is not available.", dev_name);
+			WarningAlert(str);
+			goto open_error;
+		}
+		D(bug("Real ethernet address %02x %02x %02x %02x %02x %02x\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
+
+		const char *ether_fake_address;
+		ether_fake_address = PrefsFindString("etherfakeaddress");
+		if (ether_fake_address && strlen(ether_fake_address) == 12) {
+			char sm[10];
+			strcpy( sm, "0x00" );
+			for( int i=0; i<6; i++ ) {
+				sm[2] = ether_fake_address[i*2];
+				sm[3] = ether_fake_address[i*2+1];
+				ether_addr[i] = (uint8)strtoul(sm,0,0);
+			}
+		}
+#if 1
+		/*
+		  If we bridge the underlying ethernet connection and the TAP
+		  device altogether, we have to use a fake address.
+		 */
+		else {
+			ether_addr[0] = 0x52;
+			ether_addr[1] = 0x54;
+			ether_addr[2] = 0x00;
+		}
+#endif
+		D(bug("Fake ethernet address %02x %02x %02x %02x %02x %02x\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
+	}
+	else if (net_if_type == NET_IF_SLIRP) {
+		ether_addr[0] = 0x52;
+		ether_addr[1] = 0x54;
+		ether_addr[2] = 0x00;
+		ether_addr[3] = 0x12;
+		ether_addr[4] = 0x34;
+		ether_addr[5] = 0x56;
+		D(bug("Ethernet address %02x %02x %02x %02x %02x %02x\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
+	}
+	else {
+		memcpy( ether_addr, router_mac_addr, 6 );
+		D(bug("Fake ethernet address (same as router) %02x %02x %02x %02x %02x %02x\n", ether_addr[0], ether_addr[1], ether_addr[2], ether_addr[3], ether_addr[4], ether_addr[5]));
+	}
 
 	// Start packet reception thread
 	int_ack = CreateSemaphore( 0, 0, 1, NULL);
@@ -285,27 +426,23 @@ bool ether_init(void)
 
 	ether_th = (HANDLE)_beginthreadex( 0, 0, ether_thread_feed_int, 0, 0, &ether_tid );
 	if (!ether_th) {
-		D(bug("Failed to create ethernet thread\r\n"));
+		D(bug("Failed to create ethernet thread\n"));
 		goto open_error;
 	}
 	thread_active = true;
-#if 0
-	SetThreadPriority( ether_th, threads[THREAD_ETHER].priority_running );
-	SetThreadAffinityMask( ether_th, threads[THREAD_ETHER].affinity_mask );
-#endif
 
 	unsigned int dummy;
-	ether_th2 = (HANDLE)_beginthreadex( 0, 0, ether_thread_get_packets_nt, 0, 0, &dummy );
-#if 0
-	SetThreadPriority( ether_th2, threads[THREAD_ETHER].priority_running );
-	SetThreadAffinityMask( ether_th2, threads[THREAD_ETHER].affinity_mask );
-#endif
-
+	unsigned int (WINAPI *receive_func)(void *);
+	switch (net_if_type) {
+	case NET_IF_SLIRP:
+	  receive_func = slirp_receive_func;
+	  break;
+	default:
+	  receive_func = ether_thread_get_packets_nt;
+	  break;
+	}
+	ether_th2 = (HANDLE)_beginthreadex( 0, 0, receive_func, 0, 0, &dummy );
 	ether_th1 = (HANDLE)_beginthreadex( 0, 0, ether_thread_write_packets, 0, 0, &dummy );
-#if 0
-	SetThreadPriority( ether_th1, threads[THREAD_ETHER].priority_running );
-	SetThreadAffinityMask( ether_th1, threads[THREAD_ETHER].affinity_mask );
-#endif
 
 	// Everything OK
 	return true;
@@ -328,10 +465,17 @@ bool ether_init(void)
 		int_send_now = 0;
 		thread_active = false;
 	}
-	if(!ether_fake) {
-		PacketCloseAdapter(fd);
+	if (fd) {
+		switch (net_if_type) {
+		case NET_IF_B2ETHER:
+			PacketCloseAdapter(fd);
+			break;
+		case NET_IF_TAP:
+			tap_close_adapter(fd);
+			break;
+		}
+		fd = 0;
 	}
-	fd = 0;
 	return false;
 }
 
@@ -342,45 +486,39 @@ bool ether_init(void)
 
 void ether_exit(void)
 {
-	D(bug("EtherExit\r\n"));
+	D(bug("EtherExit\n"));
 
-	// Take them down in a controlled way.
+	// Stop reception thread
 	thread_active = false;
 
-	// _asm int 3
-
-	D(bug("Closing ethernet device %s\r\n",edevice));
-
-	if(!*edevice) return;
-
 	if(int_ack) ReleaseSemaphore(int_ack,1,NULL);
 	if(int_sig) ReleaseSemaphore(int_sig,1,NULL);
 	if(int_sig2) ReleaseSemaphore(int_sig2,1,NULL);
 	if(int_send_now) ReleaseSemaphore(int_send_now,1,NULL);
 
-	D(bug("CancelIO if needed\r\n"));
+	D(bug("CancelIO if needed\n"));
 	if (fd && fd->hFile && pfnCancelIo)
 		pfnCancelIo(fd->hFile);
 
 	// Wait max 2 secs to shut down pending io. After that, kill them.
-	D(bug("Wait delay\r\n"));
+	D(bug("Wait delay\n"));
 	for( int i=0; i<10; i++ ) {
 		if(!thread_active_1 && !thread_active_2 && !thread_active_3) break;
 		Sleep(200);
 	}
 
 	if(thread_active_1) {
-		D(bug("Ether killing ether_th1\r\n"));
+		D(bug("Ether killing ether_th1\n"));
 		if(ether_th1) TerminateThread(ether_th1,0);
 		thread_active_1 = false;
 	}
 	if(thread_active_2) {
-		D(bug("Ether killing ether_th2\r\n"));
+		D(bug("Ether killing ether_th2\n"));
 		if(ether_th2) TerminateThread(ether_th2,0);
 		thread_active_2 = false;
 	}
 	if(thread_active_3) {
-		D(bug("Ether killing thread\r\n"));
+		D(bug("Ether killing thread\n"));
 		if(ether_th) TerminateThread(ether_th,0);
 		thread_active_3 = false;
 	}
@@ -389,7 +527,7 @@ void ether_exit(void)
 	ether_th2 = 0;
 	ether_th = 0;
 
-	D(bug("Closing semaphores\r\n"));
+	D(bug("Closing semaphores\n"));
 	if(int_ack) {
 		CloseHandle(int_ack);
 		int_ack = 0;
@@ -408,13 +546,20 @@ void ether_exit(void)
 	}
 
 	// Close ethernet device
-	if(fd) {
-		PacketCloseAdapter(fd);
+	if (fd) {
+		switch (net_if_type) {
+		case NET_IF_B2ETHER:
+			PacketCloseAdapter(fd);
+			break;
+		case NET_IF_TAP:
+			tap_close_adapter(fd);
+			break;
+		}
 		fd = 0;
 	}
 
 	// Remove all protocols
-	D(bug("Removing protocols\r\n"));
+	D(bug("Removing protocols\n"));
 	NetProtocol *p = prot_list;
 	while (p) {
 		NetProtocol *next = p->next;
@@ -423,35 +568,203 @@ void ether_exit(void)
 	}
 	prot_list = 0;
 
-	D(bug("Deleting sections\r\n"));
+	D(bug("Deleting sections\n"));
 	DeleteCriticalSection( &fetch_csection );
 	DeleteCriticalSection( &queue_csection );
 	DeleteCriticalSection( &send_csection );
 	DeleteCriticalSection( &wpool_csection );
 
-	D(bug("Freeing read packets\r\n"));
+	D(bug("Freeing read packets\n"));
 	free_read_packets();
 
-	D(bug("Freeing write packets\r\n"));
+	D(bug("Freeing write packets\n"));
 	free_write_packets();
 
-	D(bug("Finalizing queue\r\n"));
+	D(bug("Finalizing queue\n"));
 	final_queue();
 
-	D(bug("Stopping router\r\n"));
-	router_final();
+	if (net_if_type == NET_IF_ROUTER) {
+		D(bug("Stopping router\n"));
+		router_final();
+	}
 
-	D(bug("EtherExit done\r\n"));
+	D(bug("EtherExit done\n"));
 }
 
 
+/*
+ *  Glue around low-level implementation
+ */
+
+#ifdef SHEEPSHAVER
+// Error codes
+enum {
+	eMultiErr		= -91,
+	eLenErr			= -92,
+	lapProtErr		= -94,
+	excessCollsns	= -95
+};
+
+// Initialize ethernet
+void EtherInit(void)
+{
+	net_open = false;
+
+	// Do nothing if the user disabled the network
+	if (PrefsFindBool("nonet"))
+		return;
+
+	net_open = ether_init();
+}
+
+// Exit ethernet
+void EtherExit(void)
+{
+	ether_exit();
+	net_open = false;
+}
+
+// Get ethernet hardware address
+void AO_get_ethernet_address(uint32 arg)
+{
+	uint8 *addr = Mac2HostAddr(arg);
+	if (net_open)
+		OTCopy48BitAddress(ether_addr, addr);
+	else {
+		addr[0] = 0x12;
+		addr[1] = 0x34;
+		addr[2] = 0x56;
+		addr[3] = 0x78;
+		addr[4] = 0x9a;
+		addr[5] = 0xbc;
+	}
+	D(bug("AO_get_ethernet_address: got address %02x%02x%02x%02x%02x%02x\n", addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]));
+}
+
+// Add multicast address
+void AO_enable_multicast(uint32 addr)
+{
+	if (net_open)
+		ether_do_add_multicast(Mac2HostAddr(addr));
+}
+
+// Disable multicast address
+void AO_disable_multicast(uint32 addr)
+{
+	if (net_open)
+		ether_do_del_multicast(Mac2HostAddr(addr));
+}
+
+// Transmit one packet
+void AO_transmit_packet(uint32 mp)
+{
+	if (net_open) {
+		switch (ether_do_write(mp)) {
+		case noErr:
+			num_tx_packets++;
+			break;
+		case excessCollsns:
+			num_tx_buffer_full++;
+			break;
+		}
+	}
+}
+
+// Copy packet data from message block to linear buffer
+static inline int ether_arg_to_buffer(uint32 mp, uint8 *p)
+{
+	return ether_msgb_to_buffer(mp, p);
+}
+
+// Ethernet interrupt
+void EtherIRQ(void)
+{
+	D(bug("EtherIRQ\n"));
+	num_ether_irq++;
+
+	OTEnterInterrupt();
+	ether_do_interrupt();
+	OTLeaveInterrupt();
+
+	// Acknowledge interrupt to reception thread
+	D(bug(" EtherIRQ done\n"));
+	ReleaseSemaphore(int_ack,1,NULL);
+}
+#else
+// Add multicast address
+int16 ether_add_multicast(uint32 pb)
+{
+	return ether_do_add_multicast(Mac2HostAddr(pb + eMultiAddr));
+}
+
+// Disable multicast address
+int16 ether_del_multicast(uint32 pb)
+{
+	return ether_do_del_multicast(Mac2HostAddr(pb + eMultiAddr));
+}
+
+// Transmit one packet
+int16 ether_write(uint32 wds)
+{
+	return ether_do_write(wds);
+}
+
+// Copy packet data from WDS to linear buffer
+static inline int ether_arg_to_buffer(uint32 wds, uint8 *p)
+{
+	return ether_wds_to_buffer(wds, p);
+}
+
+// Dispatch packet to protocol handler
+static void ether_dispatch_packet(uint32 packet, uint32 length)
+{
+	// Get packet type
+	uint16 type = ReadMacInt16(packet + 12);
+
+	// Look for protocol
+	NetProtocol *prot = find_protocol(type);
+	if (prot == NULL)
+		return;
+
+	// No default handler
+	if (prot->handler == 0)
+		return;
+
+	// Copy header to RHA
+	Mac2Mac_memcpy(ether_data + ed_RHA, packet, 14);
+	D(bug(" header %08lx%04lx %08lx%04lx %04lx\n", ReadMacInt32(ether_data + ed_RHA), ReadMacInt16(ether_data + ed_RHA + 4), ReadMacInt32(ether_data + ed_RHA + 6), ReadMacInt16(ether_data + ed_RHA + 10), ReadMacInt16(ether_data + ed_RHA + 12)));
+
+	// Call protocol handler
+	M68kRegisters r;
+	r.d[0] = type;								// Packet type
+	r.d[1] = length - 14;						// Remaining packet length (without header, for ReadPacket)
+	r.a[0] = packet + 14;						// Pointer to packet (Mac address, for ReadPacket)
+	r.a[3] = ether_data + ed_RHA + 14;			// Pointer behind header in RHA
+	r.a[4] = ether_data + ed_ReadPacket;		// Pointer to ReadPacket/ReadRest routines
+	D(bug(" calling protocol handler %08lx, type %08lx, length %08lx, data %08lx, rha %08lx, read_packet %08lx\n", prot->handler, r.d[0], r.d[1], r.a[0], r.a[3], r.a[4]));
+	Execute68k(prot->handler, &r);
+}
+
+// Ethernet interrupt
+void EtherInterrupt(void)
+{
+	D(bug("EtherIRQ\n"));
+	ether_do_interrupt();
+
+	// Acknowledge interrupt to reception thread
+	D(bug(" EtherIRQ done\n"));
+	ReleaseSemaphore(int_ack,1,NULL);
+}
+#endif
+
+
 /*
  *  Reset
  */
 
 void ether_reset(void)
 {
-	D(bug("EtherReset\r\n"));
+	D(bug("EtherReset\n"));
 
 	// Remove all protocols
 	NetProtocol *p = prot_list;
@@ -468,18 +781,21 @@ void ether_reset(void)
  *  Add multicast address
  */
 
-int16 ether_add_multicast(uint32 pb)
+static int16 ether_do_add_multicast(uint8 *addr)
 {
-	D(bug("ether_add_multicast\r\n"));
+	D(bug("ether_add_multicast\n"));
 
 	// We wouldn't need to do this
 	// if(ether_multi_mode != ETHER_MULTICAST_MAC) return noErr;
 
-	if (!ether_fake && !PacketAddMulticast( fd, Mac2HostAddr(pb + eMultiAddr))) {
-		D(bug("WARNING: couldn't enable multicast address\r\n"));
-		return eMultiErr;
-	} else {
-		D(bug("ether_add_multicast: noErr\r\n"));
+	switch (net_if_type) {
+	case NET_IF_B2ETHER:
+		if (!PacketAddMulticast( fd, addr)) {
+			D(bug("WARNING: couldn't enable multicast address\n"));
+			return eMultiErr;
+		}
+	default:
+		D(bug("ether_add_multicast: noErr\n"));
 		return noErr;
 	}
 }
@@ -489,18 +805,22 @@ int16 ether_add_multicast(uint32 pb)
  *  Delete multicast address
  */
 
-int16 ether_del_multicast(uint32 pb)
+int16 ether_do_del_multicast(uint8 *addr)
 {
-	D(bug("ether_del_multicast\r\n"));
+	D(bug("ether_del_multicast\n"));
 
 	// We wouldn't need to do this
 	// if(ether_multi_mode != ETHER_MULTICAST_MAC) return noErr;
 
-	if (!ether_fake && !PacketDelMulticast( fd, Mac2HostAddr(pb + eMultiAddr))) {
-		D(bug("WARNING: couldn't disable multicast address\r\n"));
-		return eMultiErr;
-	} else
+	switch (net_if_type) {
+	case NET_IF_B2ETHER:
+		if (!PacketDelMulticast( fd, addr)) {
+			D(bug("WARNING: couldn't disable multicast address\n"));
+			return eMultiErr;
+		}
+	default:
 		return noErr;
+	}
 }
 
 
@@ -510,12 +830,12 @@ int16 ether_del_multicast(uint32 pb)
 
 int16 ether_attach_ph(uint16 type, uint32 handler)
 {
-	D(bug("ether_attach_ph type=0x%x, handler=0x%x\r\n",(int)type,handler));
+	D(bug("ether_attach_ph type=0x%x, handler=0x%x\n",(int)type,handler));
 
 	// Already attached?
 	NetProtocol *p = find_protocol(type);
 	if (p != NULL) {
-		D(bug("ether_attach_ph: lapProtErr\r\n"));
+		D(bug("ether_attach_ph: lapProtErr\n"));
 		return lapProtErr;
 	} else {
 		// No, create and attach
@@ -524,7 +844,7 @@ int16 ether_attach_ph(uint16 type, uint32 handler)
 		p->type = type;
 		p->handler = handler;
 		prot_list = p;
-		D(bug("ether_attach_ph: noErr\r\n"));
+		D(bug("ether_attach_ph: noErr\n"));
 		return noErr;
 	}
 }
@@ -536,7 +856,7 @@ int16 ether_attach_ph(uint16 type, uint32 handler)
 
 int16 ether_detach_ph(uint16 type)
 {
-	D(bug("ether_detach_ph type=%08lx\r\n",(int)type));
+	D(bug("ether_detach_ph type=%08lx\n",(int)type));
 
 	NetProtocol *p = find_protocol(type);
 	if (p != NULL) {
@@ -569,10 +889,10 @@ static void dump_packet( uint8 *packet, int length )
 	if(length > 256) length = 256;
 
 	for (int i=0; i<length; i++) {
-		sprintf(sm,"%02x", (int)packet[i]);
+		sprintf(sm," %02x", (int)packet[i]);
 		strcat( buf, sm );
 	}
-	strcat( buf, "\r\n" );
+	strcat( buf, "\n" );
 	bug(buf);
 }
 #endif
@@ -629,7 +949,7 @@ static void free_write_packets( void )
 	int i = 0;
 	while(write_packet_pool) {
 		next = write_packet_pool->next;
-		D(bug("Freeing write packet %ld\r\n",++i));
+		D(bug("Freeing write packet %ld\n",++i));
 		PacketFreePacket(write_packet_pool);
 		write_packet_pool = next;
 	}
@@ -640,7 +960,7 @@ void recycle_write_packet( LPPACKET Packet )
 	EnterCriticalSection( &wpool_csection );
 	Packet->next = write_packet_pool;
 	write_packet_pool = Packet;
-	D(bug("Pool size after recycling = %ld\r\n",get_write_packet_pool_sz()));
+	D(bug("Pool size after recycling = %ld\n",get_write_packet_pool_sz()));
 	LeaveCriticalSection( &wpool_csection );
 }
 
@@ -665,7 +985,7 @@ static LPPACKET get_write_packet( UINT len )
 		Packet = PacketAllocatePacket(fd,len);
 	}
 
-	D(bug("Pool size after get wr packet = %ld\r\n",get_write_packet_pool_sz()));
+	D(bug("Pool size after get wr packet = %ld\n",get_write_packet_pool_sz()));
 
 	LeaveCriticalSection( &wpool_csection );
 
@@ -678,25 +998,43 @@ static unsigned int ether_thread_write_packets(void *arg)
 
 	thread_active_1 = true;
 
-	D(bug("ether_thread_write_packets start\r\n"));
+	D(bug("ether_thread_write_packets start\n"));
 
 	while(thread_active) {
 		// must be alertable, otherwise write completion is never called
 		WaitForSingleObjectEx(int_send_now,INFINITE,TRUE);
 		while( thread_active && (Packet = get_send_head()) != 0 ) {
-			if(m_router_enabled && router_write_packet((uint8 *)Packet->Buffer, Packet->Length)) {
+			switch (net_if_type) {
+			case NET_IF_ROUTER:
+				if(router_write_packet((uint8 *)Packet->Buffer, Packet->Length)) {
+					Packet->bIoComplete = TRUE;
+					recycle_write_packet(Packet);
+				}
+				break;
+			case NET_IF_FAKE:
 				Packet->bIoComplete = TRUE;
 				recycle_write_packet(Packet);
-			} else if(ether_fake) {
+				break;
+			case NET_IF_B2ETHER:
+				if(!PacketSendPacket( fd, Packet, FALSE, TRUE )) {
+					// already recycled if async
+				}
+				break;
+			case NET_IF_TAP:
+				if (!tap_send_packet(fd, Packet, FALSE, TRUE)) {
+					// already recycled if async
+				}
+				break;
+			case NET_IF_SLIRP:
+				slirp_input((uint8 *)Packet->Buffer, Packet->Length);
 				Packet->bIoComplete = TRUE;
 				recycle_write_packet(Packet);
-			} else if(!PacketSendPacket( fd, Packet, FALSE, TRUE )) {
-				// already recycled if async
+				break;
 			}
 		}
 	}
 
-	D(bug("ether_thread_write_packets exit\r\n"));
+	D(bug("ether_thread_write_packets exit\n"));
 
 	thread_active_1 = false;
 
@@ -707,7 +1045,7 @@ static BOOL write_packet( uint8 *packet, int len )
 {
 	LPPACKET Packet;
 
-	D(bug("write_packet\r\n"));
+	D(bug("write_packet\n"));
 
 	Packet = get_write_packet(len);
 	if(Packet) {
@@ -728,29 +1066,16 @@ static BOOL write_packet( uint8 *packet, int len )
 	}
 }
 
-int16 ether_write(uint32 wds)
+static int16 ether_do_write(uint32 arg)
 {
-	D(bug("ether_write\r\n"));
-
-	// Set source address
-	uint32 hdr = ReadMacInt32(wds + 2);
-	memcpy(Mac2HostAddr(hdr + 6), ether_addr, 6);
+	D(bug("ether_write\n"));
 
 	// Copy packet to buffer
 	uint8 packet[1514], *p = packet;
-	int len = 0;
-	for (;;) {
-		uint16 w = (uint16)ReadMacInt16(wds);
-		if (w == 0)
-			break;
-		memcpy(p, Mac2HostAddr(ReadMacInt32(wds + 2)), w);
-		len += w;
-		p += w;
-		wds += 6;
-	}
+	int len = ether_arg_to_buffer(arg, p);
 
 	if(len > 1514) {
-		D(bug("illegal packet length: %d\r\n",len));
+		D(bug("illegal packet length: %d\n",len));
 		return eLenErr;
 	} else {
 #if MONITOR
@@ -761,7 +1086,7 @@ int16 ether_write(uint32 wds)
 
 	// Transmit packet
 	if (!write_packet(packet, len)) {
-		D(bug("WARNING: couldn't transmit packet\r\n"));
+		D(bug("WARNING: couldn't transmit packet\n"));
 		return excessCollsns;
 	} else {
 		// It's up to the protocol drivers to do the error checking. Even if the
@@ -791,11 +1116,11 @@ static void final_queue(void)
 	}
 }
 
-void enqueue_packet( uint8 *buf, int sz )
+void enqueue_packet( const uint8 *buf, int sz )
 {
 	EnterCriticalSection( &queue_csection );
 	if(queue[queue_inx].sz > 0) {
-		D(bug("ethernet queue full, packet dropped\r\n"));
+		D(bug("ethernet queue full, packet dropped\n"));
 	} else {
 		if(sz > 1514) sz = 1514;
 		queue[queue_inx].sz = sz;
@@ -832,7 +1157,7 @@ static void trigger_queue(void)
 {
 	EnterCriticalSection( &queue_csection );
 	if( queue[queue_head].sz > 0 ) {
-		D(bug(" packet received, triggering Ethernet interrupt\r\n"));
+		D(bug(" packet received, triggering Ethernet interrupt\n"));
 		SetInterruptFlag(INTFLAG_ETHER);
 		TriggerInterrupt();
 		// of course can't wait here.
@@ -855,6 +1180,233 @@ static bool set_wait_request(void)
 }
 
 
+/*
+ *  TAP-Win32 glue
+ */
+
+static LPADAPTER tap_open_adapter(const char *dev_name)
+{
+	fd = (LPADAPTER)GlobalAllocPtr(GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof(*fd));
+	if (fd == NULL)
+		return NULL;
+
+	char dev_path[MAX_PATH];
+	snprintf(dev_path, sizeof(dev_path),
+			 "\\\\.\\Global\\%s.tap", dev_name);
+
+	HANDLE handle = CreateFile(
+		dev_path,
+		GENERIC_READ | GENERIC_WRITE,
+		0,
+		NULL,
+		OPEN_EXISTING,
+		FILE_ATTRIBUTE_SYSTEM | FILE_FLAG_OVERLAPPED,
+		NULL);
+	if (handle == NULL || handle == INVALID_HANDLE_VALUE)
+		return NULL;
+
+	fd->hFile = handle;
+	return fd;
+}
+
+static void tap_close_adapter(LPADAPTER fd)
+{
+	if (fd) {
+		if (fd->hFile) {
+			tap_set_status(fd, false);
+			CloseHandle(fd->hFile);
+		}
+		GlobalFreePtr(fd);
+	}
+}
+
+static bool tap_check_version(LPADAPTER fd)
+{
+	ULONG len;
+	ULONG info[3] = { 0, };
+
+	if (!DeviceIoControl(fd->hFile, TAP_IOCTL_GET_VERSION,
+						 &info, sizeof(info),
+						 &info, sizeof(info), &len, NULL)
+		&& !DeviceIoControl(fd->hFile, OLD_TAP_IOCTL_GET_VERSION,
+							&info, sizeof(info),
+							&info, sizeof(info), &len, NULL))
+		return false;
+
+	if (info[0] > TAP_VERSION_MIN_MAJOR)
+		return true;
+	if (info[0] == TAP_VERSION_MIN_MAJOR && info[1] >= TAP_VERSION_MIN_MINOR)
+		return true;
+
+	return false;
+}
+
+static bool tap_set_status(LPADAPTER fd, ULONG status)
+{
+	DWORD len = 0;
+	return DeviceIoControl(fd->hFile, TAP_IOCTL_SET_MEDIA_STATUS,
+						   &status, sizeof (status),
+						   &status, sizeof (status), &len, NULL);
+}
+
+static bool tap_get_mac(LPADAPTER fd, LPBYTE addr)
+{
+	DWORD len = 0;
+	return DeviceIoControl(fd->hFile, TAP_IOCTL_GET_MAC,
+						   addr, 6,
+						   addr, 6, &len, NULL);
+						   
+}
+
+static VOID CALLBACK tap_write_completion(
+	DWORD dwErrorCode,
+	DWORD dwNumberOfBytesTransfered,
+	LPOVERLAPPED lpOverLapped
+	)
+{
+	LPPACKET lpPacket = CONTAINING_RECORD(lpOverLapped, PACKET, OverLapped);
+
+	lpPacket->bIoComplete = TRUE;
+	recycle_write_packet(lpPacket);
+}
+
+static bool tap_send_packet(
+	LPADAPTER fd,
+	LPPACKET lpPacket,
+	BOOLEAN Sync,
+	BOOLEAN RecyclingAllowed)
+{
+	BOOLEAN Result;
+
+	lpPacket->OverLapped.Offset = 0;
+	lpPacket->OverLapped.OffsetHigh = 0;
+	lpPacket->bIoComplete = FALSE;
+
+	if (Sync) {
+		Result = WriteFile(fd->hFile,
+						   lpPacket->Buffer,
+						   lpPacket->Length,
+						   &lpPacket->BytesReceived,
+						   &lpPacket->OverLapped);
+		if (Result) {
+			GetOverlappedResult(fd->hFile,
+								&lpPacket->OverLapped,
+								&lpPacket->BytesReceived,
+								TRUE);
+		}
+		lpPacket->bIoComplete = TRUE;
+		if (RecyclingAllowed)
+			PacketFreePacket(lpPacket);
+	}
+	else {
+		Result = WriteFileEx(fd->hFile,
+							 lpPacket->Buffer,
+							 lpPacket->Length,
+							 &lpPacket->OverLapped,
+							 tap_write_completion);
+
+		if (!Result && RecyclingAllowed)
+			recycle_write_packet(lpPacket);
+	}
+
+	return Result;
+}
+
+static bool tap_receive_packet(LPADAPTER fd, LPPACKET lpPacket, BOOLEAN Sync)
+{
+	BOOLEAN Result;
+
+	lpPacket->OverLapped.Offset = 0;
+	lpPacket->OverLapped.OffsetHigh = 0;
+	lpPacket->bIoComplete = FALSE;
+
+	if (Sync) {
+		Result = ReadFile(fd->hFile,
+						  lpPacket->Buffer,
+						  lpPacket->Length,
+						  &lpPacket->BytesReceived,
+						  &lpPacket->OverLapped);
+		if (Result) {
+			Result = GetOverlappedResult(fd->hFile,
+										 &lpPacket->OverLapped,
+										 &lpPacket->BytesReceived,
+										 TRUE);
+			if (Result)
+				lpPacket->bIoComplete = TRUE;
+			else
+				lpPacket->free = TRUE;
+		}
+	}
+	else {
+		Result = ReadFileEx(fd->hFile,
+							lpPacket->Buffer,
+							lpPacket->Length,
+							&lpPacket->OverLapped,
+							packet_read_completion);
+
+		if (!Result)
+			lpPacket->BytesReceived = 0;
+	}
+
+	return Result;
+}
+
+
+/*
+ *  SLIRP output buffer glue
+ */
+
+int slirp_can_output(void)
+{
+	return 1;
+}
+
+void slirp_output(const uint8 *packet, int len)
+{
+	enqueue_packet(packet, len);
+}
+
+static unsigned int slirp_receive_func(void *arg)
+{
+	D(bug("slirp_receive_func\n"));
+	thread_active_2 = true;
+
+	while (thread_active) {
+		// Wait for packets to arrive
+		fd_set rfds, wfds, xfds;
+		int nfds, ret, timeout;
+
+		// ... in the output queue
+		nfds = -1;
+		FD_ZERO(&rfds);
+		FD_ZERO(&wfds);
+		FD_ZERO(&xfds);
+		timeout = slirp_select_fill(&nfds, &rfds, &wfds, &xfds);
+#if ! USE_SLIRP_TIMEOUT
+		timeout = 10000;
+#endif
+		if (nfds < 0) {
+			/* Windows does not honour the timeout if there is not
+			   descriptor to wait for */
+			Delay_usec(timeout);
+			ret = 0;
+		}
+		else {
+			struct timeval tv;
+			tv.tv_sec = 0;
+			tv.tv_usec = timeout;
+			ret = select(0, &rfds, &wfds, &xfds, &tv);
+		}
+		if (ret >= 0)
+			slirp_select_poll(&rfds, &wfds, &xfds);
+	}
+
+	D(bug("slirp_receive_func exit\n"));
+	thread_active_2 = false;
+	return 0;
+}
+
+
 /*
  *  Packet reception threads
  */
@@ -881,19 +1433,24 @@ VOID CALLBACK packet_read_completion(
 				if(count == pending_packet_sz[j] &&
 				   memcmp(pending_packet[j],lpPacket->Buffer,count) == 0)
 				{
-					D(bug("packet_read_completion discarding own packet.\r\n"));
+					D(bug("packet_read_completion discarding own packet.\n"));
 					dwNumberOfBytesTransfered = 0;
 
 					j = (j+1) & (~(MAX_ECHO-1));
 					if(j != echo_count) {
-						D(bug("Wow, this fix made some good after all...\r\n"));
+						D(bug("Wow, this fix made some good after all...\n"));
 					}
 
 					break;
 				}
 			}
+			// XXX drop packets that are not for us
+			if (net_if_type == NET_IF_TAP) {
+				if (memcmp((LPBYTE)lpPacket->Buffer, ether_addr, 6) != 0)
+					dwNumberOfBytesTransfered = 0;
+			}
 			if(dwNumberOfBytesTransfered) {
-				if(!m_router_enabled || !router_read_packet((uint8 *)lpPacket->Buffer, dwNumberOfBytesTransfered)) {
+				if(net_if_type != NET_IF_ROUTER || !router_read_packet((uint8 *)lpPacket->Buffer, dwNumberOfBytesTransfered)) {
 					enqueue_packet( (LPBYTE)lpPacket->Buffer, dwNumberOfBytesTransfered );
 				}
 			}
@@ -937,7 +1494,7 @@ static bool allocate_read_packets(void)
 	for( int i=0; i<PACKET_POOL_COUNT; i++ ) {
 		packets[i] = PacketAllocatePacket(fd,1514);
 		if(!packets[i]) {
-			D(bug("allocate_read_packets: out of memory\r\n"));
+			D(bug("allocate_read_packets: out of memory\n"));
 			return(false);
 		}
 	}
@@ -958,20 +1515,29 @@ static unsigned int ether_thread_get_packets_nt(void *arg)
 
 	thread_active_2 = true;
 
-	D(bug("ether_thread_get_packets_nt start\r\n"));
+	D(bug("ether_thread_get_packets_nt start\n"));
 
 	// Wait for packets to arrive.
 	// Obey the golden rules; keep the reads pending.
 	while(thread_active) {
 
-		if(!ether_fake) {
-			D(bug("Pending reads\r\n"));
+		if(net_if_type == NET_IF_B2ETHER || net_if_type == NET_IF_TAP) {
+			D(bug("Pending reads\n"));
 			for( i=0; thread_active && i<PACKET_POOL_COUNT; i++ ) {
 				if(packets[i]->free) {
 					packets[i]->free = FALSE;
-					if(PacketReceivePacket(fd,packets[i],FALSE)) {
+					BOOLEAN Result;
+					switch (net_if_type) {
+					case NET_IF_B2ETHER:
+						Result = PacketReceivePacket(fd, packets[i], FALSE);
+						break;
+					case NET_IF_TAP:
+						Result = tap_receive_packet(fd, packets[i], FALSE);
+						break;
+					}
+					if (Result) {
 						if(packets[i]->bIoComplete) {
-							D(bug("Early io completion...\r\n"));
+							D(bug("Early io completion...\n"));
 							packet_read_completion(
 								ERROR_SUCCESS,
 								packets[i]->BytesReceived,
@@ -986,13 +1552,13 @@ static unsigned int ether_thread_get_packets_nt(void *arg)
 		}
 
 		if(thread_active && has_no_completed_io()) {
-			D(bug("Waiting for int_sig2\r\n"));
+			D(bug("Waiting for int_sig2\n"));
 			// "problem": awakens twice in a row. Fix if you increase the pool size.
 			WaitForSingleObjectEx(int_sig2,INFINITE,TRUE);
 		}
 	}
 
-	D(bug("ether_thread_get_packets_nt exit\r\n"));
+	D(bug("ether_thread_get_packets_nt exit\n"));
 
 	thread_active_2 = false;
 
@@ -1005,13 +1571,13 @@ static unsigned int ether_thread_feed_int(void *arg)
 
 	thread_active_3 = true;
 
-	D(bug("ether_thread_feed_int start\r\n"));
+	D(bug("ether_thread_feed_int start\n"));
 
 	while(thread_active) {
-		D(bug("Waiting for int_sig\r\n"));
+		D(bug("Waiting for int_sig\n"));
 		WaitForSingleObject(int_sig,INFINITE);
 		// Looping this way to avoid a race condition.
-		D(bug("Triggering\r\n"));
+		D(bug("Triggering\n"));
 		looping = true;
 		while(thread_active && looping) {
 			trigger_queue();
@@ -1019,10 +1585,10 @@ static unsigned int ether_thread_feed_int(void *arg)
 			WaitForSingleObject(int_ack,INFINITE);
 			if(thread_active) looping = set_wait_request();
 		}
-		D(bug("Queue empty.\r\n"));
+		D(bug("Queue empty.\n"));
 	}
 
-	D(bug("ether_thread_feed_int exit\r\n"));
+	D(bug("ether_thread_feed_int exit\n"));
 
 	thread_active_3 = false;
 
@@ -1034,58 +1600,27 @@ static unsigned int ether_thread_feed_int(void *arg)
  *  Ethernet interrupt - activate deferred tasks to call IODone or protocol handlers
  */
 
-void EtherInterrupt(void)
+static void ether_do_interrupt(void)
 {
-	int length;
+	// Call protocol handler for received packets
 	EthernetPacket ether_packet;
 	uint32 packet = ether_packet.addr();
+	ssize_t length;
+	for (;;) {
 
-	D(bug("EtherIRQ\r\n"));
-
-	// Call protocol handler for received packets
-	while( (length = dequeue_packet(Mac2HostAddr(packet))) > 0 ) {
-
+		// Read packet from Ethernet device
+		length = dequeue_packet(Mac2HostAddr(packet));
 		if (length < 14)
-			continue;
+			break;
 
 #if MONITOR
 		bug("Receiving Ethernet packet (%d bytes):\n",(int)length);
 		dump_packet( Mac2HostAddr(packet), length );
 #endif
 
-		// Get packet type
-		uint16 type = ReadMacInt16(packet + 12);
-
-		// Look for protocol
-		NetProtocol *prot = find_protocol(type);
-		if (prot == NULL)
-			continue;
-		// break;
-
-		// No default handler
-		if (prot->handler == 0)
-			continue;
-		// break;
-
-		// Copy header to RHA
-		Mac2Mac_memcpy(ether_data + ed_RHA, packet, 14);
-		D(bug(" header %08lx%04lx %08lx%04lx %04lx\r\n", ReadMacInt32(ether_data + ed_RHA), ReadMacInt16(ether_data + ed_RHA + 4), ReadMacInt32(ether_data + ed_RHA + 6), ReadMacInt16(ether_data + ed_RHA + 10), ReadMacInt16(ether_data + ed_RHA + 12)));
-
-		// Call protocol handler
-		M68kRegisters r;
-		r.d[0] = type;                  // Packet type
-		r.d[1] = length - 14;             // Remaining packet length (without header, for ReadPacket)
-
-		r.a[0] = packet + 14;                     // Pointer to packet (Mac address, for ReadPacket)
-		r.a[3] = ether_data + ed_RHA + 14;        // Pointer behind header in RHA
-		r.a[4] = ether_data + ed_ReadPacket;      // Pointer to ReadPacket/ReadRest routines
-		D(bug(" calling protocol handler %08lx, type %08lx, length %08lx, data %08lx, rha %08lx, read_packet %08lx\r\n", prot->handler, r.d[0], r.d[1], r.a[0], r.a[3], r.a[4]));
-		Execute68k(prot->handler, &r);
+		// Dispatch packet
+		ether_dispatch_packet(packet, length);
 	}
-
-	// Acknowledge interrupt to reception thread
-	D(bug(" EtherIRQ done\r\n"));
-	ReleaseSemaphore(int_ack,1,NULL);
 }
 
 #if DEBUG
